COMPONENT_KEY = "meta_extractor"


class Tasks:
    SINGLE = "extract_meta_from_text"
    PIPELINE = "run_meta_extractor_with_core_logic"


class Queue:
    MAIN = "meta_extractor"


class StatusKeys:
    EXTRACT_METADATA = "extract_metadata"